import logging
from datetime import datetime
from typing import Optional

import boto3
import pytz
from botocore.config import Config
from botocore.exceptions import ClientError

from timestream import get_logger, log_it

DEFAULT_WRITE_CLIENT_CONFIG = Config(
    read_timeout=20, max_pool_connections=5000, retries={"max_attempts": 10}
)


class TimestreamReader:
    def __init__(
        self,
        session: Optional[boto3.Session] = None,
        current_time_ms: Optional[int] = None,
        logger: Optional[logging.Logger] = None,
    ):
        self._session = session if session else boto3.Session()
        self._timestream_read_client = self._session.client("timestream-query")
        self._timestream_write_client = self._session.client(
            "timestream-write", config=DEFAULT_WRITE_CLIENT_CONFIG
        )
        self._logger = logger or get_logger()

        self._current_time_ms = self._dt_to_ms(datetime.now().astimezone(pytz.utc))
        if current_time_ms is not None:
            assert (
                type(current_time_ms) is int and current_time_ms > 0
            ), f"Invalid current_time_ms: {current_time_ms}. Expecting an integer greater than 0"
            self._logger.warning(f"Overriding current_time_ms with value {current_time_ms}")
            self._current_time_ms = current_time_ms

    @log_it
    def does_database_exist(
        self, database_name: Optional[str] = None, raise_err: bool = False
    ) -> bool:
        """
        Performs a boto3 describe_database API call and returns True
        if the specified database_name exists

        Parameters
        ----------
        database_name : str
            The name of the database you want to check

        raise_err : bool, Default : False
            Whether to raise the boto3 ResourceNotFoundException or just return
            False

        Returns
        -------
        bool

        Raises
        ------
        botocore.exceptions.ClientError/ResourceNotFoundException
        """
        assert database_name, "Cannot check if a database exists without a database name"
        try:
            response = self._timestream_write_client.describe_database(DatabaseName=database_name)
            result = True
        except ClientError as ex:
            if raise_err:
                raise ex
            else:
                result = False
        return result

    @log_it
    def does_table_exist(
        self,
        database_name: Optional[str] = None,
        table_name: Optional[str] = None,
        raise_err: bool = False,
    ) -> bool:
        """
        Performs a boto3 describe_table API call and returns True if the
        specified table_name exists in the database

        Parameters
        ----------
        database_name : str
            The name of the database you want to check

        table_name : str
            The name of the database you want to check

        raise_err : bool, Default : False
            Whether to raise the boto3 ResourceNotFoundException or just return
            False

        Returns
        -------
        bool

        Raises
        ------
        botocore.exceptions.ClientError/ResourceNotFoundException
        """
        assert database_name, "Cannot check if a table exists without a database name"
        assert table_name, "Cannot check if a table exists without a table name"
        try:
            response = self._timestream_write_client.describe_table(
                DatabaseName=database_name, TableName=table_name
            )
            result = True
        except ClientError as ex:
            if raise_err:
                raise ex
            else:
                result = False
        return result

    @log_it
    def custom_read(self, sql: Optional[str] = None, file_path: Optional[str] = None) -> list[dict]:
        """Executes a Timestream query from a string or file
        If both `sql` and `file_path` are provided, `file_path` takes precedent

        Parameters
        ----------
        sql : str
            SQL query string

        file_path : str
            Relative or full path to a file containing the SQL query

        Returns
        -------
        records : list[dict]
        """
        assert sql or file_path, "Must provide a query string or path"
        if file_path:
            with open(file_path, "r") as file_object:
                sql = file_object.read()

            self._logger.info(f"SQL query read from {file_path}: {sql}")

        resp = self._timestream_read_client.query(QueryString=sql, MaxRows=1000)
        next_token = resp.get("NextToken")
        rows = resp.get("Rows") or []
        columns = resp.get("ColumnInfo") or []
        while next_token:
            resp = self._timestream_read_client.query(
                QueryString=sql, NextToken=next_token, MaxRows=1000
            )
            next_token = resp.get("NextToken")
            rows.extend(resp.get("Rows") or [])
        self._logger.info(f"Retrieved {len(rows)} records from Timestream")
        return self._parse_rows(rows, columns)

    @log_it
    def read(
        self,
        database_name: str,
        table_name: str,
        projection_expression: Optional[str] = None,
        measure_value: Optional[float] = None,
        measure_value_op: Optional[float] = None,
        time: Optional[datetime] = None,
        min_time: Optional[datetime] = None,
        max_time: Optional[datetime] = None,
        limit: Optional[int] = None,
        **kwargs,
    ) -> list[dict]:
        """Generates a SQL query string based on the provided parameters executes it.
        Returns records as a list of dictionaries.

        Parameters
        ----------
        database_name : str
            (required) Name of the database to query

        table_name : str
            (required) Name of the database table to query

        projection_expression : str
            (optional) Expression to filter columns returned.
            For example:
             - t_AppID,measure_name,measure_value::double,time
             - COUNT(time)

            By default, uses '*'

        measure_value : float
            (optional) Measure value to filter records by.
            To set the operator, see "measure_value_op"

        measure_value_op : str
            (optional) Mathematical operator used when filtering by measure_value.
            One of [>, <, =, <>, >=, <=]
            If unspecified, uses '='

        time : datetime
            (optional) Timestamp to filter records by (searches for exact match).
            Forced into UTC time.
            If specified, "min_time" and "max_time" cannot be specified.

        min_time : datetime
            (optional) Timestamp to filter records by (searches for records later than this time).
            Forced into UTC time.
            If specified, "time" cannot be specified. May be used in conjunction with "max_time"

        max_time : datetime
            (optional) Timestamp to filter records by (searches for records earlier than this time).
            Forced into UTC time.
            If specified, "time" cannot be specified. May be used in conjunction with "min_time"

        limit : int
            (optional) Limit for the total number of records to return. Must be greater than 0.

        kwargs : dict
            (optional) All other keyword args are used for string comparisons. Accepts kwarg values
            of either strings or lists of strings.
            For example:
                reader._generate_sql(
                    dimensionA='first'
                    dimensionB=['second', 'third']
                )
            generates the where clause:
                dimensionA = 'first' AND dimensionB IN ('second','third')

        Returns
        -------
        records : list[dict]
        """
        sql = self._generate_sql(
            database_name,
            table_name,
            projection_expression,
            measure_value,
            measure_value_op,
            time,
            min_time,
            max_time,
            limit,
            **kwargs,
        )
        resp = self._timestream_read_client.query(QueryString=sql, MaxRows=1000)
        next_token = resp.get("NextToken")
        rows = resp.get("Rows") or []
        columns = resp.get("ColumnInfo") or []
        while next_token:
            resp = self._timestream_read_client.query(
                QueryString=sql, NextToken=next_token, MaxRows=1000
            )
            next_token = resp.get("NextToken")
            rows.extend(resp.get("Rows") or [])
        self._logger.info(f"Retrieved {len(rows)} records from Timestream")
        return self._parse_rows(rows, columns)

    @log_it
    def _generate_sql(
        self,
        database_name: str,
        table_name: str,
        projection_expression: Optional[str] = None,
        measure_value: Optional[float] = None,
        measure_value_op: Optional[float] = None,
        time: Optional[datetime] = None,
        min_time: Optional[datetime] = None,
        max_time: Optional[datetime] = None,
        limit: Optional[int] = None,
        **kwargs,
    ) -> str:
        assert database_name, f"Cannot generate SQL without a database name"
        assert table_name, f"Cannot generate SQL without a table name"
        assert projection_expression is None or isinstance(
            projection_expression, str
        ), f"Invalid projection expression: {projection_expression}. Expected a string"
        assert measure_value is None or isinstance(
            measure_value, float
        ), f"Invalid measure value: {measure_value}. Expected a float"
        assert measure_value_op in [
            None,
            ">",
            "<",
            "=",
            "<>",
            ">=",
            "<=",
        ], f"Invalid measure value op: {measure_value_op}"
        assert time is None or isinstance(
            time, datetime
        ), f"Invalid time: {time}. Expected a datetime"
        assert min_time is None or isinstance(
            min_time, datetime
        ), f"Invalid min_time: {min_time}. Expected a datetime"
        assert max_time is None or isinstance(
            max_time, datetime
        ), f"Invalid max_time: {max_time}. Expected a datetime"
        assert not (time and min_time), f"Cannot specify min_time when time is specified"
        assert not (time and max_time), f"Cannot specify max_time when time is specified"
        assert limit is None or (
            isinstance(limit, int) and limit > 0
        ), f"Invalid limit: {limit}. Expected an integer greater than 0"
        projection_expression = projection_expression or "*"
        measure_value_op = measure_value_op or "="
        time = None if not time else time.astimezone(pytz.utc)
        min_time = None if not min_time else min_time.astimezone(pytz.utc)
        max_time = None if not max_time else max_time.astimezone(pytz.utc)

        # Build SQL query
        sql = f'SELECT {projection_expression} FROM "{database_name}"."{table_name}"'
        clauses = []
        if measure_value:
            clauses.append(f"measure_value::double {measure_value_op} {measure_value}")
        if time:
            clauses.append(f"time = FROM_ISO8601_TIMESTAMP('{time.isoformat()}')")
        if max_time and min_time:
            clauses.append(
                f"time BETWEEN FROM_ISO8601_TIMESTAMP('{min_time.isoformat()}') AND FROM_ISO8601_TIMESTAMP('{max_time.isoformat()}')"
            )
        if max_time and not min_time:
            clauses.append(f"time <= FROM_ISO8601_TIMESTAMP('{max_time.isoformat()}')")
        if min_time and not max_time:
            clauses.append(f"time >= FROM_ISO8601_TIMESTAMP('{min_time.isoformat()}')")
        for attr, val in kwargs.items():
            if isinstance(val, str):
                clauses.append(f"{attr} = '{val}'")
            elif isinstance(val, list) and all(isinstance(v, str) for v in val):
                vals = "','".join(val)
                clauses.append(f"{attr} IN ('{vals}')")
            else:
                raise AssertionError(
                    f"Invalid kwarg value: {val}. Expected either a string or list of strings"
                )

        if clauses:
            clauses = " AND ".join(clauses)
            sql += f" WHERE {clauses}"
        if limit:
            sql += f" LIMIT {limit}"
        self._logger.info(f"Generated SQL query: {sql}")
        return sql

    def _parse_rows(self, rows: list[dict], columns: list[dict]) -> list[dict]:
        """Converts rows+columns returned by Timestream into a list of record dicts"""
        assert isinstance(rows, list), f"Invalid rows to parse: {rows}. Expected a list"
        assert isinstance(columns, list), f"Invalid columns to parse: {columns}. Expected a list"

        records = []
        for r in rows:
            assert isinstance(r, dict), f"Invalid row: {r}. Expected a dict"
            data = r.get("Data") or []
            assert len(data) == len(
                columns
            ), f"Invalid row data: {data}. Does not match number of columns"

            record = {}
            for i in range(len(columns)):
                assert (
                    isinstance(columns[i], dict) and "Name" in columns[i]
                ), f'Invalid column {columns[i]}. Expected a dict with the key "Name"'
                d, col_name = data[i], columns[i]["Name"]
                if d.get("ScalarValue") is not None:
                    record[col_name] = d["ScalarValue"]
                elif d.get("NullValue"):
                    record[col_name] = None
                else:
                    raise Exception(f"Unrecognized type for data: {d}")

            if "measure_value::double" in record:
                record["measure_value::double"] = float(record.pop("measure_value::double"))
            if "time" in record:
                record["time"] = datetime.strptime(
                    record.pop("time"), "%Y-%m-%d %H:%M:%S.%f000"
                ).replace(tzinfo=pytz.utc)
            records.append(record)
        return records

    def _dt_to_ms(self, dt: datetime) -> int:
        """Converts a datetime into a POSIX timestamp in milliseconds"""
        if dt:
            return int(round(dt.timestamp() * 1000))
        return -1



import pytest
import pytz

from botocore.exceptions import ClientError
from datetime import date, datetime
from timestream.timestream_reader import DEFAULT_WRITE_CLIENT_CONFIG, TimestreamReader
from test_utils.mock import MockApi


class TestTimestreamLoader_UnitTests:

    @pytest.mark.unit
    def test_does_database_exist(self):
        # Set expectations
        expected_error = ClientError(
            {
                "response": {
                    "Error": {
                        "Message": "The database default does not exist.",
                        "Code": "ResourceNotFoundException",
                    },
                    "ResponseMetadata": {
                        "RequestId": "V6E3FBONKV7C6YK3WLSVISIJGI",
                        "HTTPStatusCode": 404,
                        "HTTPHeaders": {
                            "x-amzn-requestid": "V6E3FBONKV7C6YK3WLSVISIJGI",
                            "content-type": "application/x-amz-json-1.0",
                            "content-length": "129",
                            "date": "Wed, 20 Oct 2021 17:04:42 GMT",
                        },
                        "RetryAttempts": 0,
                    },
                    "Message": "The database default does not exist.",
                },
                "operation_name": "DescribeDatabase",
            },
            "DescribeDatabase",
        )
        mock_response = {
            "Database": {
                "Arn": "arn:aws:timestream:us-east-1:000000000000:database/test",
                "DatabaseName": "test",
                "TableCount": 1,
                "KmsKeyId": "arn:aws:kms:us-east-1:000000000000:key/12345678-1234-1234-1234-1234567890ab",
                "CreationTime": datetime(2021, 7, 19, 16, 25, 26, 419000),
                "LastUpdatedTime": datetime(2021, 10, 8, 16, 25, 37, 893000),
            },
            "ResponseMetadata": {
                "RequestId": "1234567890abcdefghijklmnop",
                "HTTPStatusCode": 200,
                "HTTPHeaders": {
                    "x-amzn-requestid": "1234567890abcdefghijklmnop",
                    "content-type": "application/x-amz-json-1.0",
                    "content-length": "289",
                    "date": "Wed, 20 Oct 2021 16:54:48 GMT",
                },
                "RetryAttempts": 0,
            },
        }

        # Set up mocks
        mock_write_client = MockApi()
        mock_write_client.on(
            "describe_database",
            DatabaseName="default",
        ).raise_exception(expected_error)
        mock_write_client.on(
            "describe_database",
            DatabaseName="test",
        ).return_val(mock_response)
        mock_session = self.get_mock_session(write_client=mock_write_client)

        # Test False when using TimestreamReader's database attribute
        reader = TimestreamReader(mock_session)
        result = reader.does_database_exist("default")
        assert not result

        # Test exception raised if not found
        with pytest.raises(ClientError):
            reader.does_database_exist("default", raise_err=True)

        # Test True using test name
        result = reader.does_database_exist("test")
        assert result

        mock_session.assert_expectations()
        mock_write_client.assert_expectations()

    @pytest.mark.unit
    def test_does_table_exist(self):
        # Set expectations
        expected_error = ClientError(
            {
                "response": {
                    "Error": {
                        "Message": "The table default does not exist.",
                        "Code": "ResourceNotFoundException",
                    },
                    "ResponseMetadata": {
                        "RequestId": "V6E3FBONKV7C6YK3WLSVISIJGI",
                        "HTTPStatusCode": 404,
                        "HTTPHeaders": {
                            "x-amzn-requestid": "V6E3FBONKV7C6YK3WLSVISIJGI",
                            "content-type": "application/x-amz-json-1.0",
                            "content-length": "129",
                            "date": "Wed, 20 Oct 2021 17:04:42 GMT",
                        },
                        "RetryAttempts": 0,
                    },
                    "Message": "The table default does not exist.",
                },
                "operation_name": "DescribeTable",
            },
            "DescribeTable",
        )
        mock_response = {
            "Table": {
                "Arn": "arn:aws:timestream:us-east-1:000000000000:database/test/table/test",
                "TableName": "test",
                "DatabaseName": "test",
                "TableStatus": "ACTIVE",
                "RetentionProperties": {
                    "MemoryStoreRetentionPeriodInHours": 720,
                    "MagneticStoreRetentionPeriodInDays": 1,
                },
                "CreationTime": datetime(2021, 8, 19, 18, 21, 1, 519000),
                "LastUpdatedTime": datetime(2021, 8, 31, 12, 36, 4, 503000),
            },
            "ResponseMetadata": {
                "RequestId": "SDKOJJS7STGCQXOFVLXPBYDRKI",
                "HTTPStatusCode": 200,
                "HTTPHeaders": {
                    "x-amzn-requestid": "SDKOJJS7STGCQXOFVLXPBYDRKI",
                    "content-type": "application/x-amz-json-1.0",
                    "content-length": "366",
                    "date": "Wed, 20 Oct 2021 17:41:09 GMT",
                },
                "RetryAttempts": 1,
            },
        }

        # Set up mocks
        mock_write_client = MockApi()
        mock_write_client.on(
            "describe_table", DatabaseName="default", TableName="default"
        ).raise_exception(expected_error)
        mock_write_client.on("describe_table", DatabaseName="test", TableName="test").return_val(
            mock_response
        )
        mock_session = self.get_mock_session(write_client=mock_write_client)

        # Test False when using TimestreamReader's database and table attribute
        reader = TimestreamReader(mock_session)
        result = reader.does_table_exist("default", "default")
        assert not result

        # Test exception raised if not found
        with pytest.raises(ClientError) as e_info:
            reader.does_table_exist("default", "default", raise_err=True)

        # Test True using test name
        result = reader.does_table_exist("test", "test")
        assert result

        mock_session.assert_expectations()
        mock_write_client.assert_expectations()

    @pytest.mark.unit
    def test_custom_read(self):
        # Set expectations
        expected_time = datetime(2022, 1, 1, 11, 12, 13, 14, pytz.utc)
        expected_time_str = expected_time.strftime("%Y-%m-%d %H:%M:%S.%f000")
        expected_inline_sql = 'SELECT * FROM "db"."table" LIMIT 5'
        expected_file_sql = 'SELECT *\nFROM "db"."table"\nLIMIT 5'
        expected_rows = [
            {
                "Data": [
                    {"ScalarValue": 1.0},
                    {"NullValue": True},
                    {"ScalarValue": expected_time_str},
                ]
            }
        ]
        expected_columns = [
            {"Name": "measure_value::double"},
            {"Name": "dimension1"},
            {"Name": "time"},
        ]
        expected_next_token = "token"
        expected_resp = {
            "Rows": expected_rows,
            "ColumnInfo": expected_columns,
            "NextToken": expected_next_token,
        }
        expected_records = [
            {"measure_value::double": 1.0, "dimension1": None, "time": expected_time}
        ]

        # Set up mocks
        mock_custom_read_client = MockApi()
        mock_custom_read_client.on(
            "query", QueryString=expected_inline_sql, MaxRows=1000
        ).return_val(expected_resp)
        mock_custom_read_client.on(
            "query", QueryString=expected_inline_sql, NextToken=expected_next_token, MaxRows=1000
        ).return_val({})
        mock_custom_read_client.on("query", QueryString=expected_file_sql, MaxRows=1000).return_val(
            expected_resp
        )
        mock_custom_read_client.on(
            "query", QueryString=expected_file_sql, NextToken=expected_next_token, MaxRows=1000
        ).return_val({})

        # Test custom_read with inline SQL
        reader = TimestreamReader(self.get_mock_session(read_client=mock_custom_read_client))
        records = reader.custom_read(sql='SELECT * FROM "db"."table" LIMIT 5')
        assert records == expected_records

        # Test custom_read with SQL from file
        reader = TimestreamReader(self.get_mock_session(read_client=mock_custom_read_client))
        records = reader.custom_read(file_path="test/queries/test-custom-read.sql")
        assert records == expected_records

    @pytest.mark.unit
    def test_read(self):
        # Set expectations
        expected_time = datetime(2022, 1, 1, 11, 12, 13, 14, pytz.utc)
        expected_time_str = expected_time.strftime("%Y-%m-%d %H:%M:%S.%f000")
        expected_sql = 'SELECT * FROM "db"."table" LIMIT 5'
        expected_rows = [
            {
                "Data": [
                    {"ScalarValue": 1.0},
                    {"NullValue": True},
                    {"ScalarValue": expected_time_str},
                ]
            }
        ]
        expected_columns = [
            {"Name": "measure_value::double"},
            {"Name": "dimension1"},
            {"Name": "time"},
        ]
        expected_next_token = "token"
        expected_resp = {
            "Rows": expected_rows,
            "ColumnInfo": expected_columns,
            "NextToken": expected_next_token,
        }
        expected_records = [
            {"measure_value::double": 1.0, "dimension1": None, "time": expected_time}
        ]

        # Set up mocks
        mock_read_client = MockApi()
        mock_read_client.on("query", QueryString=expected_sql, MaxRows=1000).return_val(
            expected_resp
        )
        mock_read_client.on(
            "query", QueryString=expected_sql, NextToken=expected_next_token, MaxRows=1000
        ).return_val({})

        # Test read
        reader = TimestreamReader(self.get_mock_session(read_client=mock_read_client))
        records = reader.read("db", "table", limit=5)
        assert records == expected_records

    @pytest.mark.unit
    def test_generate_sql(self):
        reader = TimestreamReader(self.get_mock_session())
        time = datetime(2022, 1, 1, 11, 12, 13, 14, pytz.utc)
        min_time = datetime.combine(time.date(), datetime.min.time()).replace(tzinfo=pytz.utc)
        max_time = datetime.combine(time.date(), datetime.max.time()).replace(tzinfo=pytz.utc)

        # Test projection expression
        sql = reader._generate_sql("db", "table", projection_expression="COUNT(*)")
        assert sql == 'SELECT COUNT(*) FROM "db"."table"'

        # Test measure value
        sql = reader._generate_sql("db", "table", measure_value=1.0, measure_value_op="<>")
        assert sql == 'SELECT * FROM "db"."table" WHERE measure_value::double <> 1.0'

        # Test time
        sql = reader._generate_sql("db", "table", time=time)
        assert (
            sql
            == 'SELECT * FROM "db"."table" WHERE time = FROM_ISO8601_TIMESTAMP(\'2022-01-01T11:12:13.000014+00:00\')'
        )

        # Test min time
        sql = reader._generate_sql("db", "table", min_time=min_time)
        assert (
            sql
            == 'SELECT * FROM "db"."table" WHERE time >= FROM_ISO8601_TIMESTAMP(\'2022-01-01T00:00:00+00:00\')'
        )

        # Test max time
        sql = reader._generate_sql("db", "table", max_time=max_time)
        assert (
            sql
            == 'SELECT * FROM "db"."table" WHERE time <= FROM_ISO8601_TIMESTAMP(\'2022-01-01T23:59:59.999999+00:00\')'
        )

        # Test min + max time
        sql = reader._generate_sql("db", "table", min_time=min_time, max_time=max_time)
        assert (
            sql
            == "SELECT * FROM \"db\".\"table\" WHERE time BETWEEN FROM_ISO8601_TIMESTAMP('2022-01-01T00:00:00+00:00') AND FROM_ISO8601_TIMESTAMP('2022-01-01T23:59:59.999999+00:00')"
        )

        # Test kwargs + limit
        sql = reader._generate_sql(
            "db", "table", limit=5, dim1="val1", t_AppID=["SVC98765", "SVC01234"]
        )
        assert (
            sql
            == "SELECT * FROM \"db\".\"table\" WHERE dim1 = 'val1' AND t_AppID IN ('SVC98765','SVC01234') LIMIT 5"
        )

    @pytest.mark.unit
    def test_generate_sql_assertions(self):
        reader = TimestreamReader(self.get_mock_session())
        time = datetime(2022, 1, 1, 11, 12, 13, 14, pytz.utc)
        min_time = datetime.combine(time.date(), datetime.min.time()).replace(tzinfo=pytz.utc)
        max_time = datetime.combine(time.date(), datetime.max.time()).replace(tzinfo=pytz.utc)

        with pytest.raises(AssertionError) as e_info:
            reader._generate_sql(None, "table")
        assert str(e_info.value).startswith("Cannot generate SQL without a database")

        with pytest.raises(AssertionError) as e_info:
            reader._generate_sql("db", None)
        assert str(e_info.value).startswith("Cannot generate SQL without a table")

        with pytest.raises(AssertionError) as e_info:
            reader._generate_sql("db", "table", projection_expression=1)
        assert str(e_info.value).startswith("Invalid projection expression")

        with pytest.raises(AssertionError) as e_info:
            reader._generate_sql("db", "table", measure_value="a")
        assert str(e_info.value).startswith("Invalid measure value")

        with pytest.raises(AssertionError) as e_info:
            reader._generate_sql("db", "table", measure_value_op=10)
        assert str(e_info.value).startswith("Invalid measure value op")

        with pytest.raises(AssertionError) as e_info:
            reader._generate_sql("db", "table", time=10)
        assert str(e_info.value).startswith("Invalid time")

        with pytest.raises(AssertionError) as e_info:
            reader._generate_sql("db", "table", min_time=10)
        assert str(e_info.value).startswith("Invalid min_time")

        with pytest.raises(AssertionError) as e_info:
            reader._generate_sql("db", "table", max_time=10)
        assert str(e_info.value).startswith("Invalid max_time")

        with pytest.raises(AssertionError) as e_info:
            reader._generate_sql("db", "table", time=time, min_time=min_time)
        assert str(e_info.value).startswith("Cannot specify min_time when time is specified")

        with pytest.raises(AssertionError) as e_info:
            reader._generate_sql("db", "table", time=time, max_time=max_time)
        assert str(e_info.value).startswith("Cannot specify max_time when time is specified")

        with pytest.raises(AssertionError) as e_info:
            reader._generate_sql("db", "table", limit="10")
        assert str(e_info.value).startswith("Invalid limit")

        with pytest.raises(AssertionError) as e_info:
            reader._generate_sql("db", "table", limit=-1)
        assert str(e_info.value).startswith("Invalid limit")

        with pytest.raises(AssertionError) as e_info:
            reader._generate_sql("db", "table", dim1=10)
        assert str(e_info.value).startswith("Invalid kwarg value")

    @pytest.mark.unit
    def test_parse_rows(self):
        # Set expectations
        expected_time = datetime(2022, 1, 1, 11, 12, 13, 14, pytz.utc)
        expected_time_str = expected_time.strftime("%Y-%m-%d %H:%M:%S.%f000")
        expected_rows = [
            {
                "Data": [
                    {"ScalarValue": 1.0},
                    {"NullValue": True},
                    {"ScalarValue": expected_time_str},
                ]
            },
            {
                "Data": [
                    {"ScalarValue": 2.0},
                    {"Unknown": "bad"},
                    {"ScalarValue": expected_time_str},
                ]
            },
        ]
        expected_columns = [
            {"Name": "measure_value::double"},
            {"Name": "dimension1"},
            {"Name": "time"},
        ]
        expected_records = [
            {"measure_value::double": 1.0, "dimension1": None, "time": expected_time}
        ]
        reader = TimestreamReader(self.get_mock_session())

        # Test with bad row
        with pytest.raises(Exception) as e_info:
            reader._parse_rows(expected_rows, expected_columns)
        assert str(e_info.value).startswith("Unrecognized type for data")

        # Test without bad row
        expected_rows = expected_rows[:1]
        records = reader._parse_rows(expected_rows, expected_columns)
        assert records == expected_records

    @pytest.mark.unit
    def test_parse_rows_assertions(self):
        reader = TimestreamReader(self.get_mock_session())

        with pytest.raises(AssertionError) as e_info:
            reader._parse_rows("rows", [])
        assert str(e_info.value).startswith("Invalid rows to parse")

        with pytest.raises(AssertionError) as e_info:
            reader._parse_rows([], "columns")
        assert str(e_info.value).startswith("Invalid columns to parse")

        with pytest.raises(AssertionError) as e_info:
            reader._parse_rows(["row"], [])
        assert str(e_info.value).startswith("Invalid row")

        with pytest.raises(AssertionError) as e_info:
            reader._parse_rows([{"Data": []}], [{"Name": "col1"}])
        assert str(e_info.value).startswith("Invalid row data")

        with pytest.raises(AssertionError) as e_info:
            reader._parse_rows([{"Data": ["val1"]}], ["column"])
        assert str(e_info.value).startswith("Invalid column")

        with pytest.raises(AssertionError) as e_info:
            reader._parse_rows([{"Data": ["val1"]}], [{"bad": "column"}])
        assert str(e_info.value).startswith("Invalid column")

    @pytest.mark.unit
    def test_dt_to_ms(self):
        mock_session = self.get_mock_session()
        reader = TimestreamReader(mock_session)
        mock_session.assert_expectations()

        dt = datetime.combine(date(2021, 1, 1), datetime.min.time()).replace(tzinfo=pytz.UTC)
        assert reader._dt_to_ms(dt) == 1609459200000
        assert reader._dt_to_ms(None) == -1

    def get_mock_session(self, write_client=None, read_client=None) -> MockApi:
        mock_session = MockApi()
        mock_session.on(
            "client", "timestream-write", config=DEFAULT_WRITE_CLIENT_CONFIG
        ).return_val(write_client or MockApi())
        mock_session.on("client", "timestream-query").return_val(read_client or MockApi())
        return mock_session



import logging
from typing import Optional

from timestream.dimensions import TimestreamDimensions
from timestream.record import TimestreamRecord
from timestream import get_logger

DEFAULT_MAX_BATCH_SIZE = 100  # maximum set by AWS Timestream service limits
DEFAULT_MIN_COMMON_DIMENSIONS = 3
MAX_RECORDS_SIZE = 10000


class Batch:
    def __init__(
        self,
        records: Optional[list[TimestreamRecord]] = None,
        common_dimensions: Optional[TimestreamDimensions] = None,
    ):
        self.records = records or []
        self.common_dimensions = common_dimensions or TimestreamDimensions()


def batch_records(
    records: list[TimestreamRecord],
    max_batch_size: int = DEFAULT_MAX_BATCH_SIZE,
    min_common_dimensions: int = DEFAULT_MIN_COMMON_DIMENSIONS,
    logger: Optional[logging.Logger] = None,
) -> list[Batch]:
    logger = logger or get_logger()
    if min_common_dimensions == 0:
        return dimensionless_batch(records, max_batch_size)

    batches, remaining = [], records
    logger.info(f"MAX BATCH || {max_batch_size}  MIN COMMON DIM || {min_common_dimensions}")
    while len(remaining) > 0:
        logger.info(f"Remaining records to batch || {len(remaining)}")
        batch, remaining = _batch_records_helper(
            remaining, max_batch_size, min_common_dimensions, logger
        )
        batches.append(batch)
    return batches


def _batch_records_helper(
    records: list[TimestreamRecord],
    max_batch_size: int = DEFAULT_MAX_BATCH_SIZE,
    min_common_dimensions: int = DEFAULT_MIN_COMMON_DIMENSIONS,
    logger: Optional[logging.Logger] = None,
) -> tuple[Batch, list]:
    logger = logger or get_logger()
    max_batch_size = max_batch_size if not max_batch_size == None else DEFAULT_MAX_BATCH_SIZE
    min_common_dimensions = (
        min_common_dimensions
        if not min_common_dimensions == None
        else DEFAULT_MIN_COMMON_DIMENSIONS
    )

    assert max_batch_size > 0 and max_batch_size <= DEFAULT_MAX_BATCH_SIZE
    assert min_common_dimensions >= 0

    if len(records) < 2:
        return Batch(records, TimestreamDimensions()), []
    elif len(records) > MAX_RECORDS_SIZE:
        remaining = records[MAX_RECORDS_SIZE:]
        records = records[:MAX_RECORDS_SIZE]
    else:
        remaining = []

    indexes = [0]  # indexes for all the records in the batch
    common_dimensions = records[0].dimensions
    min_common_dimensions = min(min_common_dimensions, len(common_dimensions))

    for i in range(1, len(records)):
        if len(indexes) >= max_batch_size:
            break

        shared = [
            k
            for k in common_dimensions.get_dimension_keys()
            if records[i].dimensions.get_dimension(k) == common_dimensions.get_dimension(k)
        ]

        # Add the record to the batch if it has the minimum number of shared dimensions
        if len(shared) >= min_common_dimensions:
            for k in common_dimensions.get_dimension_keys():
                if k not in shared:
                    common_dimensions.remove_dimension(k)
            indexes.append(i)

    batch = Batch([records[i] for i in indexes], common_dimensions)
    remaining.extend([records[i] for i in range(len(records)) if i not in indexes])
    return batch, remaining


def dimensionless_batch(
    records: list[TimestreamRecord], max_batch_size: int = DEFAULT_MAX_BATCH_SIZE
) -> list[Batch]:
    """Batch records in groups of max_batch_size ignoring dimensions and dimension commonalities"""
    max_batch_size = max_batch_size or DEFAULT_MAX_BATCH_SIZE
    assert isinstance(max_batch_size, int), f"Invalid max batch size: {max_batch_size}"
    batches = []
    for i in range(0, len(records), max_batch_size):
        batches.append(Batch(records=records[i : i + max_batch_size]))
    return batches



def is_time_key(key: str) -> bool:
    """Returns whether or not a dictionary key is an accepted time key"""
    """Case insensitive and ignores leading/trailing whitespace"""
    return isinstance(key, str) and key.lower().strip() == "time"


def is_version_key(key: str) -> bool:
    """Returns whether or not a dictionary key is an accepted version key"""
    """Case insensitive and ignores leading/trailing whitespace"""
    return isinstance(key, str) and key.lower().strip() == "version"



import pytest
import pytz

from botocore.exceptions import ClientError
from datetime import date, datetime
from test_utils.mock import MockApi
from timestream.api import TimestreamLoader as ApiTimestreamLoader
from timestream.timestream_reader import DEFAULT_WRITE_CLIENT_CONFIG
from timestream.timestream_loader import TimestreamLoader
from timestream.dimensions import TimestreamDimensions
from timestream.load_result import TimestreamLoadResult
from timestream.record import RejectedTimestreamRecord, TimestreamRecord


class TestTimestreamLoader_UnitTests:
    @pytest.mark.unit
    def test_load(self):
        mock_session = self.get_mock_session()
        loader = TimestreamLoader("test.db", "test.table", mock_session)
        mock_session.assert_expectations()

        result = loader.load(data=[], capture_time=datetime(2021, 1, 1, 0, 0, 0, 0, pytz.UTC))
        assert type(result) == dict
        missing_keys = [
            k
            for k in ["error", "invalid_items", "accepted_records", "rejected_records"]
            if k not in result
        ]
        assert not missing_keys, f"Result {result} is missing keys {missing_keys}"
        assert type(result["invalid_items"]) == list
        assert len(result["invalid_items"]) == 0
        assert type(result["accepted_records"]) == list
        assert len(result["accepted_records"]) == 0
        assert type(result["rejected_records"]) == list
        assert len(result["rejected_records"]) == 0
        assert result["error"] == None
        assert loader.database_name == "test.db"
        assert loader.table_name == "test.table"

    @pytest.mark.unit
    def test_load_preview(self):
        mock_session = self.get_mock_session()
        loader = TimestreamLoader("test", "test", mock_session)
        mock_session.assert_expectations()

        result = loader.load_preview(
            data=[], capture_time=datetime(2021, 1, 1, 0, 0, 0, 0, pytz.UTC)
        )
        assert type(result) == dict
        missing_keys = [
            k
            for k in ["error", "invalid_items", "accepted_records", "rejected_records"]
            if k not in result
        ]
        assert not missing_keys, f"Result {result} is missing keys {missing_keys}"
        assert type(result["invalid_items"]) == list
        assert len(result["invalid_items"]) == 0
        assert type(result["accepted_records"]) == list
        assert len(result["accepted_records"]) == 0
        assert type(result["rejected_records"]) == list
        assert len(result["rejected_records"]) == 0
        assert result["error"] == None

    @pytest.mark.unit
    def test_private_load_no_data(self):
        mock_session = self.get_mock_session()
        loader = TimestreamLoader("test", "test", mock_session)
        mock_session.assert_expectations()

        result = loader._TimestreamLoader__load(
            data=[], capture_time=datetime(2021, 1, 1, 0, 0, 0, 0, pytz.UTC), preview=True
        )
        assert type(result) == TimestreamLoadResult
        assert len(result.invalid_items) == 0
        assert len(result.accepted_records) == 0
        assert len(result.rejected_records) == 0
        assert result.error == None

    @pytest.mark.unit
    def test_private_load_insert(self):
        # Test insert
        expected_source = "test soure"
        expected_reason = "test reason"
        expected_error = ClientError(
            {"RejectedRecords": [{"RecordIndex": 1, "Reason": expected_reason}]}, "test operation"
        )
        expected_records = [
            {"MeasureName": "valid", "MeasureValue": "1.0", "Dimensions": []},
            {"MeasureName": "invalid", "MeasureValue": "0.0", "Dimensions": []},
            {
                "MeasureName": "valid",
                "MeasureValue": "2.0",
                "Dimensions": [
                    {
                        "Name": "notshared",
                        "Value": "not_shared_dimension",
                        "DimensionValueType": "VARCHAR",
                    }
                ],
            },
        ]
        expected_common_dimensions = [
            {"Name": "shared", "Value": "shared_dimension", "DimensionValueType": "VARCHAR"},
            {"Name": "source", "Value": "test soure", "DimensionValueType": "VARCHAR"},
        ]
        expected_capture_time = "1609459200000"

        mock_client = MockApi()
        mock_client.on(
            "write_records",
            DatabaseName="test",
            TableName="test",
            CommonAttributes={
                "Dimensions": expected_common_dimensions,
                "MeasureValueType": "DOUBLE",
                "Time": expected_capture_time,
                "Version": 1,
            },
            Records=expected_records,
        ).raise_exception(expected_error)
        mock_session = self.get_mock_session(mock_client)

        loader = TimestreamLoader("test", "test", mock_session)
        result = loader._TimestreamLoader__load(
            data=[
                {"valid": 1.0, "invalid": 0.0, "shared": "shared_dimension"},
                {"valid": 2.0, "shared": "shared_dimension", "notshared": "not_shared_dimension"},
            ],
            capture_time=datetime(2021, 1, 1, 0, 0, 0, 0, pytz.UTC),
            source=expected_source,
            is_multi=False,
        )
        assert type(result) == TimestreamLoadResult
        assert len(result.accepted_records) == 2
        assert result.accepted_records[0].measure_name == "valid"
        assert len(result.rejected_records) == 1
        rejected_record = result.rejected_records[0]
        assert type(rejected_record) == RejectedTimestreamRecord
        assert str(rejected_record.error) == expected_reason
        assert rejected_record.record.measure_name == "invalid"
        mock_session.assert_expectations()
        mock_client.assert_expectations()

    @pytest.mark.unit
    def test_private_load_upsert(self):
        # Test upsert
        expected_capture_time = "1609459200000"
        expected_records = [
            {
                "MeasureName": "valid",
                "MeasureValue": "1.0",
                "Dimensions": [
                    {"Name": "test_dimension", "Value": "test", "DimensionValueType": "VARCHAR"}
                ],
                "Time": expected_capture_time,
            }
        ]

        mock_client = MockApi()
        mock_client.on(
            "write_records",
            DatabaseName="test",
            TableName="test",
            CommonAttributes={
                "Dimensions": [],
                "MeasureValueType": "DOUBLE",
                "Version": int(expected_capture_time),
            },
            Records=expected_records,
        ).return_val(None)
        mock_session = self.get_mock_session(mock_client)

        loader = TimestreamLoader(
            "test", "test", mock_session, current_time_ms=int(expected_capture_time)
        )
        result = loader._TimestreamLoader__load(
            data=[{"valid": 1.0, "time": int(expected_capture_time), "test_dimension": "test"}],
            capture_time=datetime(2021, 1, 1, 0, 0, 0, 0, pytz.UTC),
            upsert=True,
            is_multi=False,
        )
        assert type(result) == TimestreamLoadResult
        assert len(result.accepted_records) == 1
        assert result.accepted_records[0].measure_name == "valid"
        assert len(result.rejected_records) == 0
        mock_session.assert_expectations()
        mock_client.assert_expectations()

    @pytest.mark.unit
    def test_private_load_empty_dimension(self):
        # Test insert with empty dimension
        expected_capture_time = "1609459200000"
        expected_records = [
            {
                "MeasureName": "valid",
                "MeasureValue": "1.0",
                "Dimensions": [],
                "Time": expected_capture_time,
            }
        ]

        mock_client = MockApi()
        mock_client.on(
            "write_records",
            DatabaseName="test",
            TableName="test",
            CommonAttributes={
                "Dimensions": [],
                "MeasureValueType": "DOUBLE",
                "Version": int(expected_capture_time),
            },
            Records=expected_records,
        ).return_val(None)
        mock_session = self.get_mock_session(mock_client)

        loader = TimestreamLoader(
            "test", "test", mock_session, current_time_ms=int(expected_capture_time)
        )
        result = loader._TimestreamLoader__load(
            data=[{"valid": 1.0, "empty_dimension": "", "time": int(expected_capture_time)}],
            capture_time=datetime(2021, 1, 1, 0, 0, 0, 0, pytz.UTC),
            upsert=True,
            is_multi=False,
        )
        assert type(result) == TimestreamLoadResult
        assert len(result.accepted_records) == 1
        assert len(result.rejected_records) == 0
        mock_session.assert_expectations()
        mock_client.assert_expectations()

    @pytest.mark.unit
    def test_write_insert_records(self):
        # Set expectations
        expected_reason = "test reason"
        expected_error = ClientError(
            {"RejectedRecords": [{"RecordIndex": 1, "Reason": expected_reason}]}, "test operation"
        )
        expected_valid_record = TimestreamRecord("valid", 1.0)
        expected_invalid_record = TimestreamRecord("invalid", 0.0)
        expected_records = [
            {"MeasureName": "valid", "MeasureValue": "1.0", "Dimensions": []},
            {"MeasureName": "invalid", "MeasureValue": "0.0", "Dimensions": []},
        ]
        expected_capture_time = "1609459200000"
        expected_source = "test"

        # Set up mocks
        mock_client = MockApi()
        mock_client.on(
            "write_records",
            DatabaseName="test",
            TableName="test",
            CommonAttributes={
                "Dimensions": [],
                "MeasureValueType": "DOUBLE",
                "Time": expected_capture_time,
                "Version": 1,
            },
            Records=expected_records,
        ).raise_exception(expected_error)
        mock_session = self.get_mock_session(mock_client)

        # Test ClientError occurs during Timestream write
        loader = TimestreamLoader("test", "test", mock_session)
        accepted_records, rejected_records = loader._TimestreamLoader__write_insert_records(
            common_dimensions=TimestreamDimensions(),
            records=[expected_valid_record, expected_invalid_record],
            source=expected_source,
            capture_time=datetime(2021, 1, 1, 0, 0, 0, 0, tzinfo=pytz.UTC),
            is_multi=False,
        )
        assert len(accepted_records) == 1
        assert accepted_records[0] == expected_valid_record
        assert len(rejected_records) == 1

        rejected_record = rejected_records[0]
        assert type(rejected_record) == RejectedTimestreamRecord
        assert str(rejected_record.error) == expected_reason
        assert rejected_record.record == expected_invalid_record

        mock_session.assert_expectations()
        mock_client.assert_expectations()

    @pytest.mark.unit
    def test_write_upsert_records(self):
        # Set expectations
        expected_version = 1609459200000
        expected_reason = "test reason"
        expected_error = ClientError(
            {"RejectedRecords": [{"RecordIndex": 1, "Reason": expected_reason}]}, "test operation"
        )
        expected_valid_record = TimestreamRecord("valid", 1.0)
        expected_invalid_record = TimestreamRecord("invalid", 0.0)
        expected_records = [
            {"MeasureName": "valid", "MeasureValue": "1.0", "Dimensions": []},
            {"MeasureName": "invalid", "MeasureValue": "0.0", "Dimensions": []},
        ]
        expected_source = "test"

        # Set up mocks
        mock_client = MockApi()
        mock_client.on(
            "write_records",
            DatabaseName="test",
            TableName="test",
            CommonAttributes={
                "Dimensions": [],
                "MeasureValueType": "DOUBLE",
                "Version": expected_version,
            },
            Records=expected_records,
        ).raise_exception(expected_error)
        mock_session = self.get_mock_session(mock_client)

        # Test ClientError occurs during Timestream write
        loader = TimestreamLoader("test", "test", mock_session, current_time_ms=expected_version)
        accepted_records, rejected_records = loader._TimestreamLoader__write_upsert_records(
            common_dimensions=TimestreamDimensions(),
            records=[expected_valid_record, expected_invalid_record],
            source=expected_source,
            is_multi=False,
        )
        assert len(accepted_records) == 1
        assert accepted_records[0] == expected_valid_record
        assert len(rejected_records) == 1

        rejected_record = rejected_records[0]
        assert type(rejected_record) == RejectedTimestreamRecord
        assert str(rejected_record.error) == expected_reason
        assert rejected_record.record == expected_invalid_record

        mock_session.assert_expectations()
        mock_client.assert_expectations()

    @pytest.mark.unit
    def test_process_client_error(self):
        mock_session = self.get_mock_session()
        loader = TimestreamLoader("test", "test", mock_session)
        mock_session.assert_expectations()

        expected_reason = "test reason"
        expected_error = ClientError(
            {"RejectedRecords": [{"RecordIndex": 1, "Reason": expected_reason}]}, "test operation"
        )
        expected_valid_record = TimestreamRecord("valid", 1.0)
        expected_invalid_record = TimestreamRecord("invalid", 0.0)

        accepted_records, rejected_records = loader._TimestreamLoader__process_client_error(
            e=expected_error, records=[expected_valid_record, expected_invalid_record]
        )
        assert len(accepted_records) == 1
        assert accepted_records[0] == expected_valid_record
        assert len(rejected_records) == 1

        rejected_record = rejected_records[0]
        assert type(rejected_record) == RejectedTimestreamRecord
        assert str(rejected_record.error) == expected_reason
        assert rejected_record.record == expected_invalid_record

    @pytest.mark.unit
    def test_parse_data(self):
        mock_session = self.get_mock_session()
        loader = TimestreamLoader("test", "test", mock_session)

        # Test no data
        result = loader._TimestreamLoader__parse_data([], False, False, False)
        assert type(result) == TimestreamLoadResult
        assert str(result)
        assert len(result.invalid_items) == 0
        assert len(result.accepted_records) == 0
        assert len(result.rejected_records) == 0
        assert result.error == None

        # Test mix of valid and invalid items
        expected_valid_item = {"integer": 10, "dimension": "str"}
        expected_invalid_item = {"unparseable": datetime(2021, 1, 1, 0, 0, 0, 0, tzinfo=pytz.UTC)}
        expected_no_dimension_item = {"integer": 5}
        expected_error_msg = "Some items could not be parsed into Timestream records"

        result = loader._TimestreamLoader__parse_data(
            data=[
                expected_valid_item,
                expected_invalid_item,
                expected_no_dimension_item,
            ],
            upsert=False,
            ignore_parse_error=False,
            is_multi=False,
        )
        assert type(result) == TimestreamLoadResult
        assert len(result.invalid_items) == 2
        assert len(result.accepted_records) == 1
        assert len(result.rejected_records) == 0
        assert str(result.error) == expected_error_msg
        assert result.invalid_items[0]["item"] == expected_invalid_item
        assert result.invalid_items[1]["item"] == expected_no_dimension_item

        record = result.accepted_records[0]
        assert record and type(record) == TimestreamRecord
        assert record.measure_name == "integer"
        assert record.measure_value == 10.0
        assert record.mstime == None
        assert record.time == None

    @pytest.mark.unit
    def test_parse_item(self):
        mock_session = self.get_mock_session()
        loader = TimestreamLoader("test", "test", mock_session)
        mock_session.assert_expectations()

        # Test no item
        assert loader._TimestreamLoader__parse_item({}, False, False, False) == []

        # Test parse insert item
        expected_dimensions = TimestreamDimensions(
            {"string": "dimension", "remove this": "dimension"}
        )
        expected_dimensions.remove_dimension("remove this")
        records = loader._TimestreamLoader__parse_item(
            {"integer": 10, "float": 20.0, "list": [10, 20.0], "string": "dimension", "none": None},
            False,
            False,
            False,
        )
        assert len(records) == 3
        assert str(records[0])

        record = records[0]
        assert record and type(record) == TimestreamRecord
        assert record.measure_name == "integer"
        assert record.measure_value == 10.0
        assert record.mstime == None
        assert record.time == None
        assert record.dimensions == expected_dimensions
        assert record.dimensions.get_dimension("string") == "dimension"
        assert str(record.dimensions)

        record = records[1]
        assert record and type(record) == TimestreamRecord
        assert record.measure_name == "float"
        assert record.measure_value == 20.0
        assert record.mstime == None
        assert record.time == None
        assert record.dimensions == expected_dimensions

        record = records[2]
        assert record and type(record) == TimestreamRecord
        assert record.measure_name == "list"
        assert record.measure_value == 2.0
        assert record.mstime == None
        assert record.time == None
        assert record.dimensions == expected_dimensions

        # Test parse upsert item
        records = loader._TimestreamLoader__parse_item(
            {"Time": 1609459200000, "integer": 10, "string": "dimension"}, True, False, False
        )
        assert len(records) == 1

        record = records[0]
        assert record and type(record) == TimestreamRecord
        assert record.measure_name == "integer"
        assert record.measure_value == 10.0
        assert record.mstime == 1609459200000
        assert record.time == datetime.combine(date(2021, 1, 1), datetime.min.time()).replace(
            tzinfo=pytz.UTC
        )

        # Test parse error
        with pytest.raises(AssertionError) as e_info:
            loader._TimestreamLoader__parse_item(
                {"unparseable": {"key": "value"}, "string": "dimension"}, False, False, False
            )
        assert (
            str(e_info.value)
            == "Could not create a record or dimension from type \"<class 'dict'>\" at key unparseable"
        )

        # Test ignore parse error
        records = loader._TimestreamLoader__parse_item(
            {"unparseable": {"key": "value"}}, False, True, False
        )
        assert len(records) == 0

    @pytest.mark.unit
    def test_parse_item_empty_dimension(self):
        mock_session = self.get_mock_session()
        loader = TimestreamLoader("test", "test", mock_session)
        mock_session.assert_expectations()

        # Test parse insert item
        expected_dimensions = TimestreamDimensions({"string": "dimension"})
        records = loader._TimestreamLoader__parse_item(
            {
                "integer": 10,
                "float": 20.0,
                "list": [10, 20.0],
                "string": "dimension",
                "empty_dimension": "",
                "none": None,
            },
            False,
            False,
            False,
        )
        assert len(records) == 3
        assert str(records[0])

        record = records[0]
        assert record and type(record) == TimestreamRecord
        assert record.measure_name == "integer"
        assert record.measure_value == 10.0
        assert record.mstime == None
        assert record.time == None
        assert record.dimensions == expected_dimensions
        assert record.dimensions.get_dimension("string") == "dimension"
        assert str(record.dimensions)

        record = records[1]
        assert record and type(record) == TimestreamRecord
        assert record.measure_name == "float"
        assert record.measure_value == 20.0
        assert record.mstime == None
        assert record.time == None
        assert record.dimensions == expected_dimensions

        record = records[2]
        assert record and type(record) == TimestreamRecord
        assert record.measure_name == "list"
        assert record.measure_value == 2.0
        assert record.mstime == None
        assert record.time == None
        assert record.dimensions == expected_dimensions

    @pytest.mark.unit
    def test_does_database_exist(self):
        # Set expectations
        mock_response = {
            "Database": {
                "Arn": "arn:aws:timestream:us-east-1:000000000000:database/test",
                "DatabaseName": "test",
                "TableCount": 1,
                "KmsKeyId": "arn:aws:kms:us-east-1:000000000000:key/12345678-1234-1234-1234-1234567890ab",
                "CreationTime": datetime(2021, 7, 19, 16, 25, 26, 419000),
                "LastUpdatedTime": datetime(2021, 10, 8, 16, 25, 37, 893000),
            },
            "ResponseMetadata": {
                "RequestId": "1234567890abcdefghijklmnop",
                "HTTPStatusCode": 200,
                "HTTPHeaders": {
                    "x-amzn-requestid": "1234567890abcdefghijklmnop",
                    "content-type": "application/x-amz-json-1.0",
                    "content-length": "289",
                    "date": "Wed, 20 Oct 2021 16:54:48 GMT",
                },
                "RetryAttempts": 0,
            },
        }

        # Set up mocks
        mock_client = MockApi()
        mock_client.on(
            "describe_database",
            DatabaseName="test",
        ).return_val(mock_response)
        mock_session = self.get_mock_session(mock_client)

        # Test True when using TimestreamLoader's database attribute
        loader = TimestreamLoader("test", "test", mock_session)
        result = loader.does_database_exist()
        assert result
        mock_session.assert_expectations()
        mock_client.assert_expectations()

    @pytest.mark.unit
    def test_does_table_exist(self):
        # Set expectations
        mock_response = {
            "Table": {
                "Arn": "arn:aws:timestream:us-east-1:000000000000:database/test/table/test",
                "TableName": "test",
                "DatabaseName": "test",
                "TableStatus": "ACTIVE",
                "RetentionProperties": {
                    "MemoryStoreRetentionPeriodInHours": 720,
                    "MagneticStoreRetentionPeriodInDays": 1,
                },
                "CreationTime": datetime(2021, 8, 19, 18, 21, 1, 519000),
                "LastUpdatedTime": datetime(2021, 8, 31, 12, 36, 4, 503000),
            },
            "ResponseMetadata": {
                "RequestId": "SDKOJJS7STGCQXOFVLXPBYDRKI",
                "HTTPStatusCode": 200,
                "HTTPHeaders": {
                    "x-amzn-requestid": "SDKOJJS7STGCQXOFVLXPBYDRKI",
                    "content-type": "application/x-amz-json-1.0",
                    "content-length": "366",
                    "date": "Wed, 20 Oct 2021 17:41:09 GMT",
                },
                "RetryAttempts": 1,
            },
        }

        # Set up mocks
        mock_client = MockApi()
        mock_client.on("describe_table", DatabaseName="test", TableName="test").return_val(
            mock_response
        )
        mock_session = self.get_mock_session(mock_client)

        # Test True when using TimestreamLoader's database and table attribute
        loader = TimestreamLoader("test", "test", mock_session)
        result = loader.does_table_exist()
        assert result
        mock_session.assert_expectations()
        mock_client.assert_expectations()

    def get_mock_session(self, write_client=None) -> MockApi:
        mock_session = MockApi()
        mock_session.on(
            "client", "timestream-write", config=DEFAULT_WRITE_CLIENT_CONFIG
        ).return_val(write_client)
        mock_session.on("client", "timestream-query").return_val(MockApi())
        return mock_session



import json
from typing import Optional


class TimestreamDimensions:
    """
    TimestreamDimensions is a class wrapping an underlying dictionary of Timestream dimensions.
    The underlying dictionary simply maps dimensions keys (strings) to dimensions values (strings).

    These objects may optionally be initialized with a dimensions dictionary, and the underlying dict
    can be updated with the "add_dimension" and "remove_dimension" functions. This class controls
    validation of the initial dimensions dictionary and all updates to that dictionary.

    This class also includes utilities for converting itself into the appropriate structure needed
    for a Timestream write, converting itself into a human friendly dict or string, and implements
    equality and a hash string functions for quick comparisons of Timestream dimensions.
    """

    def __init__(self, dimensions: Optional[dict] = None):
        assert (
            not dimensions or type(dimensions) is dict
        ), f'Invalid type for initializing TimestreamDimensions "{type(dimensions)}"'
        assert not dimensions or all(
            isinstance(k, str) and isinstance(v, str) for k, v in dimensions.items()
        ), f"All keys and values of dimensions must be strings"
        self.__dimensions = {}
        for k, v in (dimensions or {}).items():
            self.add_dimension(k, v)

    def add_dimension(self, key: str, value: str, overwrite: bool = False):
        assert key and type(key) is str, f'Cannot add dimension {key} with type "{type(key)}"'
        assert type(value) is str, f'Cannot add dimension value {value} with type "{type(value)}"'
        assert (
            overwrite or key not in self.__dimensions
        ), f"Dimension {key}:{value} already exists and cannot be added to dimensions {self}"
        self.__dimensions[key.strip()] = value.strip()

    def remove_dimension(self, key: str):
        assert key in self.__dimensions, f"Cannot remove dimension {key} from {self}"
        self.__dimensions.pop(key)

    def get_dimension(self, key: str) -> str:
        return self.__dimensions.get(key)

    def get_dimension_keys(self) -> list[str]:
        return list(self.__dimensions.keys())

    def loadable(self, omit_dimensions: Optional[list[str]] = None) -> list[dict]:
        """loadable returns a list of dictionaries that is ready to load into Timestream"""
        omit_dimensions = omit_dimensions or []
        assert isinstance(omit_dimensions, list)
        return [
            {"Name": k, "Value": v, "DimensionValueType": "VARCHAR"}
            for k, v in self.__dimensions.items()
            if k not in omit_dimensions
        ]

    def to_dict(self) -> dict:
        return self.__dimensions.copy()

    def copy(self):
        return TimestreamDimensions(self.__dimensions)

    def hashstr(self) -> str:
        """
        Provides a hash string for this TimestreamDimensions object.

        A TimestreamDimensions object is defined by its underlying dict of dimension
        keys to dimension values (irrespective of key order in the dict).

        The hash string is created by converting the underlying dict into a list of key-value
        pairs (e.g. key="vpd", value="9.1" yields "vpd‽‽‽9.1"), sorting that list alphabetically,
        and converting the list into a string using the JSON encoder.
        """
        hash_separator = "‽‽‽"
        hash_lst = [f"{k}{hash_separator}{v}" for k, v in self.__dimensions.items()]
        hash_lst.sort()
        return json.dumps(hash_lst)

    def __eq__(self, other):
        return type(self) == type(other) and self.hashstr() == other.hashstr()

    def __repr__(self) -> str:
        return json.dumps(self.to_dict())

    def __len__(self) -> int:
        return len(self.__dimensions)



import pytest

from timestream.batch import _batch_records_helper, Batch, dimensionless_batch
from timestream.dimensions import TimestreamDimensions
from timestream.record import TimestreamRecord


class TestBatch_UnitTests:

    @pytest.mark.unit
    def test_batch_records(self):
        assert True

    @pytest.mark.unit
    def test_batch_records_helper(self):
        # Test max batch size assertions
        with pytest.raises(AssertionError):
            _batch_records_helper([], max_batch_size=0)
        with pytest.raises(AssertionError):
            _batch_records_helper([], max_batch_size=101)

        # Test min common dimensions assertions
        with pytest.raises(AssertionError):
            _batch_records_helper([], min_common_dimensions=-1)

        # Test no records
        batch, remaining = _batch_records_helper([])
        assert isinstance(remaining, list)
        assert len(remaining) == 0
        assert isinstance(batch, Batch)
        assert isinstance(batch.records, list)
        assert len(batch.records) == 0

        record_a = TimestreamRecord(
            "metric",
            2,
            TimestreamDimensions(
                {"shared0": "0", "shared1": "1", "label": "a", "notshared0": "n/a"}
            ),
        )
        record_b = TimestreamRecord(
            "metric",
            3,
            TimestreamDimensions(
                {"shared0": "0", "shared1": "1", "label": "b", "notshared1": "n/a"}
            ),
        )
        record_c = TimestreamRecord(
            "metric", 4, TimestreamDimensions({"shared0": "0", "shared1": "1", "label": "c"})
        )
        record_d = TimestreamRecord(
            "metric", 5, TimestreamDimensions({"shared0": "0", "shared1": "1", "label": "d"})
        )
        records = [record_a, record_b, record_c, record_d]

        # Test batching min common dimensions 0
        batch, remaining = _batch_records_helper(records, min_common_dimensions=0)
        assert len(batch.records) == 4
        assert len(batch.common_dimensions) == 2
        assert len(remaining) == 0

        # Test batching min common dimensions 0, max batch size 2
        batch, remaining = _batch_records_helper(records, max_batch_size=2, min_common_dimensions=0)
        assert len(batch.records) == 2
        assert batch.records[0] == record_a
        assert batch.records[1] == record_b
        assert len(batch.common_dimensions) == 2
        assert len(remaining) == 2

        # Test batching min common dimensions 5
        batch, remaining = _batch_records_helper(records, min_common_dimensions=5)
        assert len(batch.records) == 1
        assert batch.records[0] == record_a
        assert len(batch.common_dimensions) == 4
        assert len(remaining) == 3

        # Test batching min common dimensions 2
        batch, remaining = _batch_records_helper(records, min_common_dimensions=2)
        assert batch.records == records
        assert len(batch.common_dimensions) == 2
        assert batch.common_dimensions.get_dimension("shared0") == "0"
        assert batch.common_dimensions.get_dimension("shared1") == "1"

    @pytest.mark.unit
    def test_dimensionless_batch(self):
        record_a = TimestreamRecord("metric", 0)
        record_b = TimestreamRecord("metric", 1)
        record_c = TimestreamRecord("metric", 2)
        record_d = TimestreamRecord("metric", 3)
        records = [record_a, record_b, record_c, record_d]

        batches = dimensionless_batch(records, max_batch_size=1)
        assert len(batches) == 4
        assert batches[0].records == [record_a]
        assert batches[1].records == [record_b]
        assert batches[2].records == [record_c]
        assert batches[3].records == [record_d]

        batches = dimensionless_batch(records, max_batch_size=2)
        assert len(batches) == 2
        assert batches[0].records == [record_a, record_b]
        assert batches[1].records == [record_c, record_d]

        batches = dimensionless_batch(records, max_batch_size=3)
        assert len(batches) == 2
        assert batches[0].records == [record_a, record_b, record_c]
        assert batches[1].records == [record_d]

        batches = dimensionless_batch(records, max_batch_size=4)
        assert len(batches) == 1
        assert batches[0].records == records

        batches = dimensionless_batch(records, max_batch_size=5)
        assert len(batches) == 1
        assert batches[0].records == records



import json
from timestream.multi_measure import TimestreamRecordMulti
from timestream.record import RejectedTimestreamRecord, TimestreamRecord
from typing import List


class TimestreamLoadResult:
    """TimestreamLoadResult is an ease of use class that wraps the results of an attempted Timestream load"""

    def __init__(self):
        self.error = None
        self.__invalid_items = []  # data items that could not be parsed into records
        self.__accepted_records = []  # records accepted by Timestream
        self.__rejected_records = []  # records rejected by Timestream and their errors

    @property
    def invalid_items(self) -> List[dict]:
        return self.__invalid_items

    @property
    def accepted_records(self) -> List[TimestreamRecord]:
        return self.__accepted_records

    @property
    def rejected_records(self) -> List[RejectedTimestreamRecord]:
        return self.__rejected_records

    def add_invalid_item(self, item: dict, error: Exception):
        assert error, "Cannot add invalid item to a Timestream load result without an error"
        self.__invalid_items.append({"item": item, "error": str(error)})

    def add_accepted_record(self, record: TimestreamRecord):
        assert (
            record and type(record) is TimestreamRecord or TimestreamRecordMulti
        ), f'Cannot add an accepted record to a Timestream load result of type "{type(record)}"'
        self.__accepted_records.append(record)

    def add_accepted_records(self, records: List[TimestreamRecord]):
        assert (
            type(records) is list
        ), f'Cannot add accepted records to a Timestream load result. Expected a list, got type "{type(records)}"'
        for r in records:
            self.add_accepted_record(r)

    def add_rejected_record(self, record: RejectedTimestreamRecord):
        assert (
            record and type(record) is RejectedTimestreamRecord
        ), f'Cannot add a rejected record to a Timestream load result of type "{type(record)}"'
        self.__rejected_records.append(record)

    def add_rejected_records(self, records: List[RejectedTimestreamRecord]):
        assert (
            type(records) is list
        ), f'Cannot add rejected records to a Timestream load result. Expected a list, got type "{type(records)}"'
        for r in records:
            self.add_rejected_record(r)

    def to_dict(self) -> dict:
        return {
            "error": str(self.error) if self.error else None,
            "invalid_items": self.__invalid_items,
            "accepted_records": [r.to_dict() for r in self.__accepted_records],
            "rejected_records": [r.to_dict() for r in self.__rejected_records],
        }

    def __repr__(self) -> str:
        return json.dumps(self.to_dict())
